# Copyright (c) 2012 David Christian

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial
# portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
# OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# ---------------------------------------------------------------------------------------------------

# Conway grid processing and cell collision logic

# This class processes conway life rules for any multidimensional array
# passed to it. Note that the cell array must match the dimensions specified
# when the object is initialised - grid width and height. In the array a zero
# represents a dead cell and any non-zero value indicates a live cell.

import random

class ConwayGrid:

    def __init__(self, grid_width, grid_height, grid_pixel_size):
        self.grid_width = grid_width
        self.grid_height = grid_height
        self.grid_pixel_size = grid_pixel_size    

    def clear_grid(self, grid):
        for y in range(1, self.grid_height - 1):
            for x in range(1, self.grid_width - 1):
                grid[x][y] = 0    

    def process_life_grid(self, life_grid, temp):
        # Pass one - in this pass we just count the neighbours of each cell
        for y in range(1, self.grid_height - 1):
            for x in range(1, self.grid_width - 1):                                    
                temp[x][y] = 0
                
                if life_grid[x-1][y-1] > 0:
                    temp[x][y] += 1
                if life_grid[x][y-1] > 0:
                    temp[x][y] += 1
                if life_grid[x+1][y-1] > 0:
                    temp[x][y] += 1
                    
                if life_grid[x-1][y] > 0:
                    temp[x][y] += 1
                if life_grid[x+1][y] > 0:
                    temp[x][y] += 1

                if life_grid[x-1][y+1] > 0:
                    temp[x][y] += 1
                if life_grid[x][y+1] > 0:
                    temp[x][y] += 1                    
                if life_grid[x+1][y+1] > 0:
                    temp[x][y] += 1

        # Pass two... now set each cell based on its current neighbour info
        for y in range(1, self.grid_height - 1):
            for x in range(1, self.grid_width - 1):                                    
                if temp[x][y] <2 or temp[x][y] > 3 :
                    life_grid[x][y] = 0

                # TODO: Don't think this is needed... double check
                #if grid[x][y] > 0 and (grid_temp[x][y] == 2 or grid_temp[x][y] == 3):                
                #    grid[x][y] = 1

                if life_grid[x][y] == 0 and temp[x][y] == 3:
                    life_grid[x][y] = 1
                    
    def process_grid_collisions(self, grid_a, grid_a_temp, grid_b, grid_b_temp):
        # The temps must be initialized by the 'process_life_grid' function
        # before this method is called for it to work properly
        collision = False        
        for y in range(1, self.grid_height - 1):
            for x in range(1, self.grid_width - 1):
                if grid_a[x][y] > 0 and grid_b[x][y] > 0:
                    collision = True
                    a_total = 0
                    b_total = 0

                    # Which has the most surrounding cells? (Which is strongest)
                    if grid_a_temp[x][y] > grid_b_temp[x][y]:
                        grid_b[x][y] = 0                    
                    elif grid_a_temp[x][y] < grid_b_temp[x][y]:
                        grid_a[x][y] = 0                        
                    else:
                        # stalemate... the cells destroy each other                        
                        grid_a[x][y] = 0
                        grid_b[x][y] = 0         
        return collision

    # Conway virus grid population logic (such as it is!)

    # This was supposed to be more comprehensive, but I'm all out of time with
    # regards to the PyWeek competition so it's greatly simplified.
    # Literally just past random shapes over the board until
    # the cell allocation is spent. Not the grand AI I have been
    # planning all week :)
    def populate_virus_grid(self, grid, cell_allocation, shape):
        while cell_allocation > 0:
            rx = random.randint(0, self.grid_width - 6)
            ry = random.randint(0, self.grid_height - 6)
            for y in range(0, 5):
                for x in range(0, 5):
                    if shape[x][y] > 0:
                        if rx + x < self.grid_width and ry + y < self.grid_height:                        
                            grid[rx + x][ry + y] = 1
                            cell_allocation -= 1
